﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace HeuristicLab.Problems.DataAnalysis.Tests {
  [TestClass]
  public class AutoDiffIntervalTest {
    private readonly AlgebraicInterval a = new AlgebraicInterval(-1, 1);
    private readonly AlgebraicInterval b = new AlgebraicInterval(-2, 2);
    private readonly AlgebraicInterval c = new AlgebraicInterval(0, 3);
    private readonly AlgebraicInterval d = new AlgebraicInterval(1, 3);
    private readonly AlgebraicInterval e = new AlgebraicInterval(4, 6);
    private readonly IntervalEvaluator eval = new IntervalEvaluator();

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalAddOperation() {
      //add        [x1,x2] + [y1,y2] = [x1 + y1,x2 + y2]

      // [-1,1] + [-2,2] = [-3,3]
      AssertAreEqualInterval(Add(a, b), new AlgebraicInterval(-3, 3));
      //([-1, 1] + [-2, 2]) + [0, 3] = [-3, 6]
      AssertAreEqualInterval(Add(Add(a, b), c), new AlgebraicInterval(-3, 6));
      //([-1, 1] + [0, 3]) + [-2, 2] = [-3, 6]
      AssertAreEqualInterval(Add(Add(a, c), b), new AlgebraicInterval(-3, 6));
    }

    private void AssertAreEqualInterval(AlgebraicInterval a, AlgebraicInterval b) {
      if (double.IsNaN(a.LowerBound.Value)) {
        Assert.IsTrue(double.IsNaN(b.LowerBound.Value));
      } else {
        Assert.AreEqual(a.LowerBound.Value.Value, b.LowerBound.Value.Value, Math.Abs(a.LowerBound.Value.Value) * 1e-4); // relative error < 0.1%
      }

      if (double.IsNaN(a.UpperBound.Value)) {
        Assert.IsTrue(double.IsNaN(b.UpperBound.Value));
      } else {
        Assert.AreEqual(a.UpperBound.Value.Value, b.UpperBound.Value.Value, Math.Abs(a.UpperBound.Value.Value) * 1e-4);
      }
    }

    private static AlgebraicInterval Add(AlgebraicInterval a, AlgebraicInterval b) {
      return a.Clone().Add(b);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalSubOperation() {
      //subtract   [x1,x2] − [y1,y2] = [x1 − y2,x2 − y1]

      //[-1, 1] - [-2, 2] = [-3, 3]
      AssertAreEqualInterval(Subtract(a, b), new AlgebraicInterval(-3, 3));
      //([-1, 1] - [-2, 2]) - [0, 3] = [-6, 3]
      AssertAreEqualInterval(Subtract(Subtract(a, b), c), new AlgebraicInterval(-6, 3));
      //([-1, 1] - [0, 3]) - [-2, 2] = [-6, 3]
      AssertAreEqualInterval(Subtract(Subtract(a, c), b), new AlgebraicInterval(-6, 3));
    }

    private AlgebraicInterval Subtract(AlgebraicInterval a, AlgebraicInterval b) {
      return a.Clone().Sub(b);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalMutlipyOperation() {
      //multiply   [x1,x2] * [y1,y2] = [min(x1*y1,x1*y2,x2*y1,x2*y2),max(x1*y1,x1*y2,x2*y1,x2*y2)]

      //[-1, 1] * [-2, 2] = [-2, 2]
      AssertAreEqualInterval(Multiply(a, b), new AlgebraicInterval(-2, 2));
      //([-1, 1] * [-2, 2]) * [0, 3] = [-6, 6]
      AssertAreEqualInterval(Multiply(Multiply(a, b), c), new AlgebraicInterval(-6, 6));
      //([-1, 1] * [0, 3]) * [-2, 2] = [-6, 6]
      AssertAreEqualInterval(Multiply(Multiply(a, c), b), new AlgebraicInterval(-6, 6));

      // [-2, 0] * [-2, 0]  = [0, 4]
      AssertAreEqualInterval(new AlgebraicInterval(0, 4), Multiply(new AlgebraicInterval(-2, 0), new AlgebraicInterval(-2, 0)));
    }

    private AlgebraicInterval Multiply(AlgebraicInterval a, AlgebraicInterval b) {
      return a.Clone().Mul(b);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalDivideOperation() {
      //divide  [x1, x2] / [y1, y2] = [x1, x2] * (1/[y1, y2]), where 1 / [y1,y2] = [1 / y2,1 / y1] if 0 not in [y_1, y_2].

      //[4, 6] / [1, 3] = [4/3, 6]
      AssertAreEqualInterval(Divide(e, d), new AlgebraicInterval(4.0 / 3.0, 6));
      //([4, 6] / [1, 3]) / [1, 3] = [4/9, 6]
      AssertAreEqualInterval(Divide(Divide(e, d), d), new AlgebraicInterval(4.0 / 9.0, 6));
      //[4, 6] / [0, 3] = [4/3, +Inf]
      AssertAreEqualInterval(Divide(e, c), new AlgebraicInterval(4.0 / 3.0, double.PositiveInfinity));
      //[-1, 1] / [0, 3] = [+Inf, -Inf]
      AssertAreEqualInterval(Divide(a, c), new AlgebraicInterval(double.NegativeInfinity, double.PositiveInfinity));

      //Division by 0 ==> IsInfiniteOrUndefined == true
      AssertAreEqualInterval(Divide(d, b), new AlgebraicInterval(double.NegativeInfinity, double.PositiveInfinity));

      // [0, 1] / [0, 1]
      AssertAreEqualInterval(new AlgebraicInterval(0, 1).Div(new AlgebraicInterval(0, 1)), new AlgebraicInterval(1, double.PositiveInfinity)); // returns [NaN, PositiveInfinity]

    }

    private AlgebraicInterval Divide(AlgebraicInterval e, AlgebraicInterval d) {
      return e.Clone().Div(d);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalSineOperator() {
      //sine depends on interval
      //sin([0, 2*pi]) = [-1, 1]
      AssertAreEqualInterval(Sine(new AlgebraicInterval(0, 2 * Math.PI)), new AlgebraicInterval(-1, 1));
      //sin([-pi/2, pi/2]) = [sin(-pi/2), sin(pi/2)]
      AssertAreEqualInterval(Sine(new AlgebraicInterval(-1 * Math.PI / 2, Math.PI / 2)), new AlgebraicInterval(-1, 1));
      //sin([0, pi/2]) = [sin(0), sin(pi/2)]
      AssertAreEqualInterval(Sine(new AlgebraicInterval(0, Math.PI / 2)), new AlgebraicInterval(0, 1));
      //sin([pi, 3*pi/2]) = [sin(pi), sin(3*pi/2)]
      AssertAreEqualInterval(Sine(new AlgebraicInterval(Math.PI, 3 * Math.PI / 2)), new AlgebraicInterval(-1, 0));
      AssertAreEqualInterval(Sine(new AlgebraicInterval(1, 2)), new AlgebraicInterval(Math.Min(Math.Sin(1), Math.Sin(2)), 1));
      AssertAreEqualInterval(Sine(new AlgebraicInterval(1, 3)), new AlgebraicInterval(Math.Min(Math.Sin(1), Math.Sin(3)), 1));
      AssertAreEqualInterval(Sine(new AlgebraicInterval(Math.PI, 5 * Math.PI / 2)), new AlgebraicInterval(-1, 1));
    }

    private AlgebraicInterval Sine(AlgebraicInterval algebraicInterval) {
      return algebraicInterval.Sin();
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalCosineOperator() {
      //Cosine uses sine Sine(Subtract(a, new AlgebraicInterval(Math.PI / 2, Math.PI / 2)));
      AssertAreEqualInterval(Cosine(new AlgebraicInterval(0, 2 * Math.PI)), new AlgebraicInterval(-1, 1));
      AssertAreEqualInterval(new AlgebraicInterval(-1, 1), Cosine(new AlgebraicInterval(Math.PI, 4 * Math.PI / 2)));
    }

    private AlgebraicInterval Cosine(AlgebraicInterval algebraicInterval) {
      return algebraicInterval.Cos();
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalLogOperator() {
      //Log([3, 5]) = [log(3), log(5)]
      AssertAreEqualInterval(new AlgebraicInterval(Math.Log(3), Math.Log(5)), Logarithm(new AlgebraicInterval(3, 5)));
      //Log([0.5, 1]) = [log(0.5), log(1)]
      AssertAreEqualInterval(new AlgebraicInterval(Math.Log(0.5), 0), Logarithm(new AlgebraicInterval(0.5, 1)));
      //Log([-1, 5]) = [NaN, log(5)]
      var result = Logarithm(new AlgebraicInterval(-1, 5));
      AssertAreEqualInterval(new AlgebraicInterval(double.NaN, Math.Log(5)), result);

    }

    private AlgebraicInterval Logarithm(AlgebraicInterval algebraicInterval) {
      return algebraicInterval.Log();
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalExpOperator() {
      //Exp([0, 1]) = [exp(0), exp(1)]
      AssertAreEqualInterval(new AlgebraicInterval(1, Math.Exp(1)), Exponential(new AlgebraicInterval(0, 1)));
    }

    private AlgebraicInterval Exponential(AlgebraicInterval algebraicInterval) {
      return algebraicInterval.Exp();
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalSqrOperator() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 4), new AlgebraicInterval(1, 2).IntPower(2));
      AssertAreEqualInterval(new AlgebraicInterval(1, 4), new AlgebraicInterval(-2, -1).IntPower(2));
      AssertAreEqualInterval(new AlgebraicInterval(0, 4), new AlgebraicInterval(-2, 2).IntPower(2));

      AssertAreEqualInterval(new AlgebraicInterval(0, 16), new AlgebraicInterval(-2, 2).Mul(new AlgebraicInterval(2, 2)).IntPower(2));
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalSqrtOperator() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 2), new AlgebraicInterval(1, 4).IntRoot(2));
      AssertAreEqualInterval(new AlgebraicInterval(double.NaN, double.NaN), new AlgebraicInterval(-4, -1).IntRoot(2));
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalCubeOperator() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 8), new AlgebraicInterval(1, 2).IntPower(3));
      AssertAreEqualInterval(new AlgebraicInterval(-8, -1), new AlgebraicInterval(-2, -1).IntPower(3));
      AssertAreEqualInterval(new AlgebraicInterval(-8, 8), new AlgebraicInterval(-2, 2).IntPower(3));
    }


    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalAbsOperator() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 2), new AlgebraicInterval(1, 2).Abs());
      AssertAreEqualInterval(new AlgebraicInterval(1, 2), new AlgebraicInterval(-2, -1).Abs());
      AssertAreEqualInterval(new AlgebraicInterval(0, 2), new AlgebraicInterval(-2, 2).Abs());
      AssertAreEqualInterval(new AlgebraicInterval(0, double.PositiveInfinity), new AlgebraicInterval(double.NegativeInfinity, double.PositiveInfinity).Abs());
      AssertAreEqualInterval(new AlgebraicInterval(double.NaN, 1), new AlgebraicInterval(-1, double.NaN).Abs()); // XXX unclear how we should handle NaN
      AssertAreEqualInterval(new AlgebraicInterval(double.NaN, 1), new AlgebraicInterval(double.NaN, -1).Abs());
    }


    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalCbrtOperator() {
      AssertAreEqualInterval(new AlgebraicInterval(1, 2), new AlgebraicInterval(1, 8).IntRoot(3));
      AssertAreEqualInterval(new AlgebraicInterval(-2, -1), new AlgebraicInterval(-8, -1).IntRoot(3));
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalAddAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("x + y");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(1, 2) },
        { "y", new Interval(0, 1) }
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(1, r.LowerBound);
      Assert.AreEqual(3, r.UpperBound);

      Assert.AreEqual(1.0, lg[0]); // x
      Assert.AreEqual(2.0, ug[0]); 
      Assert.AreEqual(0.0, lg[1]); // y
      Assert.AreEqual(1.0, ug[1]);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalMulAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("x * y");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(1, 2) },
        { "y", new Interval(0, 1) }
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(0, r.LowerBound);
      Assert.AreEqual(2, r.UpperBound);

      Assert.AreEqual(0.0, lg[0]); // x
      Assert.AreEqual(2.0, ug[0]);
      Assert.AreEqual(0.0, lg[1]); // y
      Assert.AreEqual(2.0, ug[1]);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalSqrAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("sqr(x)");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(1, 2) },
        { "y", new Interval(0, 1) }
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(XXX, r.LowerBound);
      Assert.AreEqual(XXX, r.UpperBound);

      Assert.AreEqual(XXX, lg[0]); // x
      Assert.AreEqual(XXX, ug[0]);

      for  { "x", new Interval(1, 2) },
        { "y", new Interval(0, 1) },

      0 <> -2,50012500572888E-05 for y in SQR(LOG('y'))
      0 <> 2, 49987500573946E-05 for x in SQR(LOG('x'))
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalExpAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("exp(x)");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(1, 2) },
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(Math.Exp(1), r.LowerBound);
      Assert.AreEqual(Math.Exp(2), r.UpperBound);

      Assert.AreEqual(Math.Exp(1), lg[0]); // x
      Assert.AreEqual(Math.Exp(2) * 2, ug[0]);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalSinAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("sin(x)");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(1, 2) },
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(Math.Sin(1), r.LowerBound); // sin(1) < sin(2)
      Assert.AreEqual(1, r.UpperBound); //  1..2 crosses pi / 2 and sin(pi/2)==1

      Assert.AreEqual(Math.Cos(1), lg[0]); // x
      Assert.AreEqual(0, ug[0]);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalCosAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("cos(x)");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(3, 4) },
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(-1, r.LowerBound); //  3..4 crosses pi and cos(pi) == -1
      Assert.AreEqual(Math.Cos(4), r.UpperBound); // cos(3) < cos(4)

      Assert.AreEqual(0, lg[0]); // x
      Assert.AreEqual(-4*Math.Sin(4), ug[0]);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalSqrtAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("sqrt(x)");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(4, 9) },
        { "y", new Interval(1, 2) },
        { "z", new Interval(0, 1) },
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(2, r.LowerBound); 
      Assert.AreEqual(3, r.UpperBound); 

      Assert.AreEqual(1.0, lg[0]); // x
      Assert.AreEqual(1.5, ug[0]);

      t = parser.Parse("sqrt(y)");
      paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      r = eval.Evaluate(t, intervals, paramNodes, out lg, out  ug);
      Assert.AreEqual(1, r.LowerBound);
      Assert.AreEqual(Math.Sqrt(2), r.UpperBound);

      Assert.AreEqual(0.5, lg[0]); // y
      Assert.AreEqual(0.5*Math.Sqrt(2), ug[0], 1e-5);

      t = parser.Parse("sqrt(z)");
      paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      r = eval.Evaluate(t, intervals, paramNodes, out lg, out ug);
      Assert.AreEqual(0, r.LowerBound);
      Assert.AreEqual(1, r.UpperBound);

      Assert.AreEqual(0, lg[0]); // z
      Assert.AreEqual(0.5 * Math.Sqrt(2), ug[0], 1e-5);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalCqrtAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("cuberoot(x)");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(8, 27) },
        { "y", new Interval(1, 2) },
        { "z", new Interval(0, 1) },
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(2, r.LowerBound);
      Assert.AreEqual(3, r.UpperBound);

      Assert.AreEqual(0.0, lg[0]); // x
      Assert.AreEqual(0.0, ug[0]); XXXX

      t = parser.Parse("sqrt(y)");
      paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      r = eval.Evaluate(t, intervals, paramNodes, out lg, out ug);
      Assert.AreEqual(0.0, r.LowerBound);
      Assert.AreEqual(0.0, r.UpperBound);

      Assert.AreEqual(0.0, lg[0]); // y
      Assert.AreEqual(0.0, ug[0], 1e-5);

      t = parser.Parse("sqrt(z)");
      paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      r = eval.Evaluate(t, intervals, paramNodes, out lg, out ug);
      Assert.AreEqual(0.0, r.LowerBound);
      Assert.AreEqual(0.0, r.UpperBound);

      Assert.AreEqual(0.0, lg[0]); // z
      Assert.AreEqual(0.0, ug[0], 1e-5);
    }

    [TestMethod]
    [TestCategory("Problems.DataAnalysis")]
    [TestProperty("Time", "short")]
    public void TestIntervalLogAutoDiff() {
      var eval = new IntervalEvaluator();
      var parser = new InfixExpressionParser();
      var t = parser.Parse("log(4*x)");
      var paramNodes = t.IterateNodesPostfix().Where(n => n.SubtreeCount == 0).ToArray();
      var intervals = new Dictionary<string, Interval>() {
        { "x", new Interval(1, 2) },
      };
      var r = eval.Evaluate(t, intervals, paramNodes, out double[] lg, out double[] ug);
      Assert.AreEqual(Math.Log(4), r.LowerBound);
      Assert.AreEqual(Math.Log(8), r.UpperBound);

      Assert.AreEqual(0.25, lg[0]); // x
      Assert.AreEqual(0.25, ug[0]);

    }
  }
}
